#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
===========
nodeFactory
===========

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

This module provides the functionality to map syntactic constructs to Nodes.
"""

import re

SINGLE_LINE_TYPES = list('/!|*{')
#-- start for testing only -----------------
#vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
# SINGLE_LINE_TYPES = list('/!|*>')
#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

LEGAL_EL_NAME = re.compile('^[a-zA-Z:_%][a-zA-Z0-9-:_]*$')

#----------------------------------------
class NodeFactoryError(Exception):
    pass
#----------------------------------------

class NodeFactory(object):

    def __init__(self, nodesModule=None):
        if nodesModule:
            nm = nodesModule
        else:
            import baseWriterNodes
            nm = baseWriterNodes
        self.nodesModule = nm

    def isSingleLineNode(self, lineTypeChar):
        return lineTypeChar in SINGLE_LINE_TYPES
    
    def isValidElementName(self, name):
        return LEGAL_EL_NAME.match(name)

    def mkRootNode(self):
        return self.nodesModule.Node()
     
    def mkTextNode(self, txt, enableInlineParsing=True):
        return self.nodesModule.Text(txt)

    def mkSingleLineNode(self, lineStr):
        lineStr = lineStr.lstrip() #better slow than sorry
        typeChar = lineStr[0]
        specStr = lineStr[1:].strip()
        if typeChar == '/':
            return self.nodesModule.BumlComment(txt=specStr)
        if typeChar == '!':
            return self.nodesModule.XmlComment(txt=specStr)
        if typeChar == '|':
            return self.nodesModule.DocTypeDecl(txt=specStr)
        if typeChar == '*':
            return self.nodesModule.Macro(txt=specStr)
        if typeChar == '{':
            return self.nodesModule.Template(tag=specStr[0], txt=specStr[1:].strip())
        #-- start for testing only -----------------
        #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
        #if typeChar == '>':
        #    return self.nodesModule.XmlComment(txt=specStr)
        #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        raise(NodeFactoryError('unexpeted single line node type'))
    
    def mkElementNode(self, tagName, attDict=None):
        attDict = attDict or {}
        char = tagName[0]
        if char == '?':
            return self.nodesModule.ProcessingInstruction(tag=tagName[1:].strip(), \
                                                          atts=attDict)
        if char == '[':  #no strip to preserve whitespace
            return self.nodesModule.CDATA(tag='', txt=tagName[1:])
        if self.isValidElementName(tagName):
            return self.nodesModule.Element(tag=tagName, atts=attDict)
        else:
            raise(NodeFactoryError('unexpeted element name "%s"' % tagName))
            
    



from utilities import warning, addrNameFromString
from inlineParser import parse as inlineParse

try:
    import bud.nospam
except:
    pass

try:
    import nonEx
except:
    pass

#----------

def applyInlineMarkup(startNode, mappingSpec):
    "runs the inlineParser to mark up all special inline text spans"
    allTextNodes = startNode.findAllNodes('Text')
    for tn in allTextNodes:
        newChildren = inlineParse(tn.txt, mappingSpec)
        tn.replaceWith(newChildren)

def convertToDiv(startNode, nodeFactory):
    "converts elements to DIV, putting the former tag name as class attribute"
    allEls = startNode.findAllEls()
    for el in allEls:
        el.replaceWith(nodeFactory.mkElementNode('div', {'class': el.tag}))

if 'bud' in globals(): #bud.nospam import worked
    def _protectedEmail(addrStr):
        """parses address string and returns a protected string
           expected address string is either
              'user@server.domain'   or
              'full name <user@server.domain>'
        """
        addr, name = addrNameFromString(addrStr)
        return bud.nospam.js_obfuscated_mailto(addr, name)

    def protectEmailNode(elNode):
        "converts the TextNode under elNode to anti-spam format"
        assert elNode[0].__class__.__name__ == 'Text'
        elNode[0].txt = _protectedEmail(elNode[0].txt)




from __future__ import absolute_import
from ..buml.baseWriterNodes import *
import unittest




#== Tests ========================================
class UtilityTests(unittest.TestCase):        
    "test baseWriterNodes"

#    def testNoseFindsMe(self):
#        "testNoseFindsMe: baseWriterNodes"
#        self.assertTrue(False)

    def testElementRendering(self):                    
        """various aspects of element rendering"""
        e = Element('div')
        self.assertEqual(e.singleAttStr('key', ['val']), 'key="val"')
        self.assertEqual(e.singleAttStr('key', ['v1', 'v2']), 'key="v1 v2"')
        #--
        e = Element('div')
        self.assertEqual(e.startTag(), '<div>')
        e.addAttr('class', 'cool')
        self.assertEqual(e.startTag(), '<div class="cool">')
        e.addAttr('class', 'wet')
        self.assertEqual(e.startTag(), '<div class="cool wet">')
        e.addAttr('id', 'z1')
        self.assertEqual(e.startTag(), '<div class="cool wet" id="z1">')





from __future__ import absolute_import
from ..buml.utilities import *
from ..buml.transformUtils import _protectedEmail

import re

#def testNoseFindsMe():
#    "utilities: testNoseFindsMe"
#    assert 0


def test_findPatternPos():
    "utilities: findPatternPos"
    pat = re.compile("xx")
    assert findPatternPos('aasdf', pat) == -1
    assert findPatternPos('aaxxf', pat) == 2 
    assert findPatternPos('xxsdf', pat) == 0

def test_emailObfuscation():
    "transfromUtils: email protection"
    DEBUG = False
    try:
        import bud.nospam
    except:
        pass
    if not 'bud' in locals():
        print "bud.nospam is not installed"
        print 'try: "pip install bud.nospam" in order to pass test'
        assert 0
    s = 'user@domain.org'
    ss = _protectedEmail(s)
    spec = ('<script type="text/javascript">document.write(\n'
            '"<n uers=\\"znvygb:hfre\\100qbznva\\056bet\\">'
            'hfre\\100qbznva\\056bet<\\057n>".replace'
            '(/[a-zA-Z]/g, function(c){return String.fromCharCode'
            '((c<="Z"?90:122)>=(c=c.charCodeAt(0)+13)?c:c-26);})\n'
            ');\n</script>')
    print s
    ll=ss.split('\n')
    l = spec.split('\n')
    z = zip(ll, l)
    for x, y in z:
        print x
        print y
        print
    if DEBUG:
        html = "<html><body>%s</body><html>" % ss
        open('emailNoName.html','w').write(html)
    assert ss == spec

    s = 'john dipper <user@domain.org>'
    ss = _protectedEmail(s)
    spec = ('<script type="text/javascript">document.write(\n'
            '"<n uers=\\"znvygb:hfre\\100qbznva\\056bet\\">'
            'wbua qvccre<\\057n>".replace(/[a-zA-Z]/g, '
            'function(c){return String.fromCharCode'
            '((c<="Z"?90:122)>='
            '(c=c.charCodeAt(0)+13)?c:c-26);})\n'
            ');\n</script>')
    ll=ss.split('\n')
    l = spec.split('\n')
    z = zip(ll, l)
    for x, y in z:
        print x
        print y
        print
    if DEBUG:
        html = "<html><body>%s</body><html>" % ss
        open('emailWName.html','w').write(html)
    assert ss == spec




#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
============
inlineParser
============

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

This is the standard inline parser that comes with buml.

The inlineFactory provides functions that map a parsed string to a node
(node-level function) or text (text-level )
"""
from mappingSpec import MappingSpec, MappingSpecItem
from utilities import addrNameFromString
from inlineParser import parse


class InlineMapper(object):

    def __init__(self, nodeFactoryInstance=None):
        if nodeFactoryInstance:
            nf = nodeFactoryInstance           
        else:
            import nodeFactory
            nf = nodeFactory.NodeFactory()
        self.nodeFactory = nf

    def textHandler(self, txt):
        return [self.nodeFactory.mkTextNode(txt)]

    def _singleElHandler(self, txt, elTag, elAtts=None, recurse=True):
        atts = elAtts or {}
        el = self.nodeFactory.mkElementNode(elTag, atts)
        if recurse:
            children = parse(txt, self.mappingSpec()) #recursive parsing
        else:
            children = [self.nodeFactory.mkTextNode(txt)]
        el.setChildren(children)
        return [el]

    def emHandler(self, txt):
        return self._singleElHandler(txt, 'em')

    def strongHandler(self, txt):
        return self._singleElHandler(txt, 'strong')

    def linkHandler(self, txt):
        addr, name = addrNameFromString(txt)
        attD = {'href': addr}
        return self._singleElHandler(name, 'a', elAtts=attD)

    def replaceHandler(self, txt):
        return self._singleElHandler(txt, 'replace', recurse=False)
        
    def mappingSpec(self):
        "returns the mappingSpec"
        m = MappingSpec()
        #necessary Null Mapper for Text without matches:
        m.append(MappingSpecItem('', '', self.textHandler))
        m.append(MappingSpecItem('*', '*', self.emHandler))
        m.append(MappingSpecItem('**', '**', self.strongHandler))
        m.append(MappingSpecItem('[[', ']]', self.linkHandler))
        m.append(MappingSpecItem('::', '::', self.replaceHandler, 
                                 singleWord=True))
        return m
    





from __future__ import absolute_import
from ..buml.baseNodes import *
import unittest

#def testNoseFindsMe():
#    "baseNodes: testNoseFindsMe"
#    assert 0

def test_various():
    "baseNodes: various"
    # to improve coverage
    n = Node()
    print n
    assert n.__repr__() == "<Node instance: I'm the root>"
    e = GenEl('el')
    print e
    assert e.__repr__() == "<el GenEl>"
    e.addAttr('x', 'X')
    print e.__repr__()
    assert e.__repr__() == "<el GenEl: {'x': ['X']}>"
    class MyEl(GenEl):
        pass
    me = MyEl('I')
    assert me.__repr__() == "<I MyEl>"

    #--
    m1 = MultiDict({'x': 'X'})
    m2 = MultiDict({'x': 'X'})
    assert m1 == m2
    #--
    n.addChild(e)
    print n.asNestedList()
    #assert n.asNestedList() == [n, [[e]]]
    assert n.asNestedList() == [n, [e]]
    e2 = GenEl('e2')
    n.addChild(e2)
    print n.asNestedList()
    assert n.asNestedList() == [n, [e, e2]]
    e3 = GenEl('e3')
    e2.addChild(e3)
    print n.asNestedList()
    assert n.asNestedList() == [n, [e, [e2, [e3]]]]

def test_GenElConstruction():
    "baseNodes: GenEl with dict or MultiDict as atts"
    #--dict
    e = GenEl('e', atts={'x':'X'})
    print e.attDict()
    assert e.attDict() == {'x': ['X']}
    #--MultiDict
    md = MultiDict({'x':'X'})
    e = GenEl('e', atts=md)
    assert e.attDict() == {'x': ['X']}
    #--empty
    e = GenEl('e')
    assert e.attDict() == {}

def test_replaceWith():
    "baseNodes: replaceWith"
    #replace with a single node
    root = Node()
    e1 = GenEl('e1')
    root.addChild(e1)
    e1a = GenEl('e1a')
    e1b = GenEl('e1b')
    e1.addChild(e1a)
    e1.addChild(e1b)
    e2 = GenEl('e2')
    root.addChild(e2)
    sn = GenEl('substNode')
    e1.replaceWith(sn)
    print root.pprint()
    assert root.children[0] == sn
    assert root.children[1] == e2
    assert sn.parent == root
    assert sn.children == [e1a, e1b]
    assert e1a.parent == sn
    assert e1b.parent == sn
    #replace with a list of nodes
    sn1 = GenEl('substN1')
    sn2 = GenEl('substN2')
    sn3 = GenEl('substN3')
    sn.replaceWith([sn1, sn2, sn3])
    print root.pprint()
    assert root.children[0] == sn1
    assert root.children[1] == sn2
    assert root.children[2] == sn3
    assert root.children[3] == e2
    assert sn1.parent == root
    assert sn2.parent == root
    assert sn3.parent == root
    assert sn1.children == [e1a, e1b]
    assert e1a.parent == sn1
    assert e1b.parent == sn1



#== Tests ========================================
class UtilityTests(unittest.TestCase):        

#    def testNoseFindsMe(self):
#        "testNoseFindsMe: baseNodes"
#        self.assertTrue(False)

    def sentinel(self):
        "use to see if nosetest picks it up"
        self.assertTrue(False)
    
    def testMultiDict2(self):                    
        """MultiDict2"""
        d = MultiDict()
        self.assertFalse(bool(d))
        d['x'] = 'X'
        self.assertTrue(bool(d))

    def testMultiDict(self):                    
        """MultiDict"""
        d = MultiDict()
        self.assertEqual(d.asDict(), {})
        d.insert('a', 'A')
        d.insert('b', 'B')
        self.assertEqual(d.asDict(), {'a': ['A'], 'b': ['B']})
        d.insert('b', 'B2')
        self.assertEqual(d.asDict(), {'a': ['A'], 'b': ['B', 'B2']})
        self.assertEqual(d.items(), [('a', ['A']), ('b', ['B', 'B2'])])
        dd = MultiDict()
        dd.insert('a', 'A')
        dd.insert('a', 'A2')
        dd.insert('c', 'C')
        d.update(dd)
        self.assertEqual(d.asDict(), {'a': ['A', 'A', 'A2'], 
                                  'b': ['B', 'B2'],
                                  'c': ['C']})
        #-- from dict
        ddd = MultiDict({'x': 'X', 'y': 'Y'})
        self.assertEqual(ddd.asDict(), {'x': ['X'], 'y': ['Y']})
        #-- cloning
        md = MultiDict({'hello': 'goodbye'})
        md1 = md.clone()
        self.assertEqual(md1.asDict(), md.asDict())
        md1.insert("other", "something")
        self.assertNotEqual(md1.asDict(), md.asDict())
        #-- contains
        md = MultiDict({'x': 'X', 'y': 'Y'})
        md.insert('x', 'X2')
        self.assertTrue(md.contains({'x':'X2'}))
        self.assertTrue(md.contains({'x':'X2', 'y': 'Y'}))
        self.assertFalse(md.contains({'x':'X2', 'z': 'Z'}))
        #-- __getitem__, __setitem__
        md = MultiDict({'x': 'X', 'y': 'Y'})
        self.assertEqual(md['x'], ['X'])
        md['x'] = 'A'
        self.assertEqual(md['x'], ['X', 'A'])

    def testLastNonTextNode(self):                    
        """lastNonTextNode"""
        root = GenEl('root')
        first = GenEl('first')
        second = GenEl('second')
        third = GenEl('third')
        root.addChild(first)
        first.addChild(second)
        second.addChild(third)
        self.assertEqual(root.lastNonTextNode(), third)
        third.addChild(Text('hello'))
        self.assertEqual(root.lastNonTextNode(), third)

    def testLinkNodes(self):                    
        """linkNodess"""
        root = Node()
        a = Node()
        b = Node()
        c = Node()
        l = [root, a, b, c]
        linkNodes(l)
        self.assertEqual(root.children, [a])
        self.assertEqual(a.children, [b])
        self.assertEqual(b.children, [c])
        e1 = GenEl('e1')
        e2 = GenEl('e2')
        e3 = GenEl('e3')
        e4 = GenEl('e4')
        t = Text('hello')
        linkNodes([e1, e2, e3, e4, t])
        self.assertEqual(e1.lastNonTextNode(), e4)

    def testGenEl(self):                    
        """GenEl"""
        e=GenEl('e')
        e.addAttr('k', 'v1')
        e.addAttr('k', 'v2')
        e.addAttr('l', 'x')
        self.assertEqual(e.atts.items(), [('k', ['v1', 'v2']), ('l', ['x'])])

    def testGenEl(self):                    
        """GenEl things"""
        e=GenEl() #tagless
        self.assertEqual(e.__repr__(), '<GenEl>')
        e=GenEl('tag') 
        self.assertEqual(e.__repr__(), '<tag GenEl>')
        e.atts['key'] = 'value'
        self.assertEqual(e.__repr__(), "<tag GenEl: {'key': ['value']}>")
        





#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
===============
baseWriterNodes
===============

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

This module extends the Nodes defined in baseNodes with methods to write them
out to a text (xml, html, etc.).  

The separation of this functionality into a separate module allows for easy
customization of writers through subclassing or the use of alternative writers.  
The default behavior of Node.xml() is to do pretty printing of the content.
This inserts whitespace.  Should this not be desired, it is possible to avoid
pretty printing and whitespace by calling node.xml(indenWidth=0, joinChar='')

Use config* utility functions to configure the writer after import
"""

from copy import deepcopy
import baseNodes
from utilities import warning 

#== config and config functions ===========================================
PRETTY_INDENT_WIDTH = 2
DEFAULT_MAX_SINGLE_LINE_LEN = 40
#=================================================================

class Node(baseNodes.Node):
    #abstract node should render as empty string

    #-- configuration --------
    _EMPTY_EL_CLOSURE_STR = '/'
    _JOIN_CHAR = '\n'
    _INDENT_WIDTH = PRETTY_INDENT_WIDTH
    _MAX_SINGLE_LINE_LEN = DEFAULT_MAX_SINGLE_LINE_LEN

    @classmethod
    def setEmptyElementMode(cls, mode=''):
        if mode not in ['xml', 'sgml', 'compat']:
            warning("mode must be one of ['xml', 'sgml', 'compat']")
        elif mode == 'xml':
            Node._EMPTY_EL_CLOSURE_STR = '/'
        elif mode == 'sgml':
            Node._EMPTY_EL_CLOSURE_STR = ''
        elif mode == 'compat':
            Node._EMPTY_EL_CLOSURE_STR = ' /'

    @classmethod
    def setRenderMode(cls, mode='', prettyIndentWidth=PRETTY_INDENT_WIDTH):
        if mode not in ['pretty', 'noWhiteSpace']:
            warning("mode must be one of ['pretty', 'noWhiteSpace']")
        elif mode == 'pretty':
            Node._JOIN_CHAR = '\n'
            Node._INDENT_WIDTH = prettyIndentWidth
        elif mode == 'noWhiteSpace':
            Node._JOIN_CHAR = ''
            Node._INDENT_WIDTH = 0

    @classmethod
    def setMaxSingleLineLen(cls, maxlen):
        #  How long can a single line text be to be kept 
        #  on same line with end tag
        Node._MAX_SINGLE_LINE_LEN = maxlen
    #-- end config ------

    def startTag(self):
        return '' # specialize in subclasses

    def endTag(self):
        return '' # specialize in subclasses

    def emptyTag(self):
        return '' # specialize in subclasses where needed

    def xmlLchildren(self):
        l = []
        l.append(self.startTag())
        l.append([c.xmlL() for c in self.children])
        l.append(self.endTag())
        return l

    def xmlLtxt(self):
        l = ["%s%s%s" % (self.startTag(), self.txt, self.endTag())]
        if self.children:
            l.append([c.xmlL() for c in self.children])
        return l

    def xmlLempty(self):
        return [self.emptyTag()]

    def xmlL(self):
        "returns a list of strings"
        if self.txt: #even if it has children! Must come first!
            return self.xmlLtxt()
        if self.children:
            return self.xmlLchildren()
        return self.xmlLempty()

    def indentedStr(self, str, indentLevel):
        # //2 since two lists wrappers per step
        indentStr = indentLevel * self._INDENT_WIDTH // 2 * ' '
        l = []
        for line in str.split('\n'):
            l.append('%s%s' % (indentStr, line))
        return "\n".join(l)

    def prettyXmlL(self, xmlL, indentLevel):
        l = []
        for el in xmlL:
            if type(el) == type([]):
                l.append(self.prettyXmlL(el, indentLevel+1))
            else: # a string
                l.append(self.indentedStr(el, indentLevel))
        return self._JOIN_CHAR.join(l)

    def xml(self, initialIndentLevel=-2):
        "initialIndentLevel -1 since there is a 'silent' Node as root"
        clone = deepcopy(self)
        #short text content rendered on single line
        txtNodes = clone.findAllNodes('Text')
        for n in txtNodes:
            if not '\n' in n.txt \
               and len(n.txt) < self._MAX_SINGLE_LINE_LEN \
               and len(n.parent.children) == 1:
                n.parent.txt = n.txt
                n.parent.children = ''
        #no whitespace in CDATA
        cdataNodes = clone.findAllNodes('CDATA')
        for n in cdataNodes:
            if n.children:
                n.txt = n.children[0].txt
                n.children = []
        #remove BumlComments:
        comments = clone.findAllNodes('BumlComment')
        for n in comments:
            n.parent.children.remove(n)
        return self.prettyXmlL(clone.xmlL(), initialIndentLevel).strip()


class BumlComment(baseNodes.GenEl, Node):
    pass

class Element(baseNodes.GenEl, Node):

    def singleAttStr(self, key, valL):
        valStr = ' '.join(valL)
        return '%s="%s"' % (key, valStr)

    def attStr(self):
        attL = [self.singleAttStr(k, v) for k, v in self.atts.items()]
        return ' '.join(attL)

    def tagStr(self):
        l = [self.tag]
        attStr = self.attStr()
        if attStr:  l.append(attStr)
        return ' '.join(l)

    def startTag(self):
        return '<%s>' % self.tagStr()

    def endTag(self):
        return '</%s>' % self.tag

    def emptyTag(self):
        return '<%s%s>' % (self.tagStr(), self._EMPTY_EL_CLOSURE_STR)

class ProcessingInstruction(Element):

    def emptyTag(self):
        return '<?%s?>' % self.tagStr()

    def startTag(self):
        return '<?%s ' % self.tagStr()

    def endTag(self):
        return ' ?>'

class Text(baseNodes.Text, Node):
    pass

class XmlComment(baseNodes.GenEl, Node):
    def startTag(self):
        return '<!-- '

    def endTag(self):
        return ' -->'

class DocTypeDecl(baseNodes.GenEl, Node):
    def startTag(self):
        return '<!'

    def endTag(self):
        return '>'

class CDATA(baseNodes.GenEl, Node):
    def startTag(self):
        return '<![CDATA['

    def endTag(self):
        return ']]>'

class Template(baseNodes.GenEl, Node):
    def startTag(self):
        return '{%s ' % self.tag

    def endTag(self):
        if self.tag == '%':
            return ' %}'
        elif self.tag == '{':
            return ' }}'
        else:
            warning('Unknown Template Type %s for "%s"' % (self.tag, self.txt))
            #trying anyhow:
            return ' %s}' % self.tag

class Macro(baseNodes.GenEl, Node):
    def startTag(self):
        return '*['

    def endTag(self):
        return ']*'




from __future__ import absolute_import
from ..buml.inlineMapper import *


#def testNoseFindsMe():
#    "inlineMapper: testNoseFindsMe"
#    assert 0

def test_all():
    "inlineMapper: all"
    #just to exercise some code
    im = InlineMapper()
    assert len(im.mappingSpec()) == 5

def test_handlers():
    "inlineMapper: handlers"
    im = InlineMapper()
    t = im.textHandler('hello')
    assert t[0].__class__.__name__ == 'Text'
    assert t[0].txt == 'hello'
    em = im.emHandler('important')
    assert em[0].__class__.__name__ == 'Element'
    assert em[0].tag == 'em'
    assert em[0][0].__class__.__name__ == 'Text'
    assert em[0][0].txt == 'important'
    lnk = im.linkHandler('some name <http://bruegger.it>')
    assert lnk[0].__class__.__name__ == 'Element'
    assert lnk[0].tag == 'a'
    assert lnk[0].atts['href'] == ['http://bruegger.it']
    assert lnk[0][0].__class__.__name__ == 'Text'
    print '"%s"' % lnk[0][0].txt
    assert lnk[0][0].txt == 'some name'




def warning(msgStr):
    print "Warning: %s" % msgStr

def findPatternPos(inStr, pattern):
    "for re what find is for string: returns pos of first separator or -1"
    match = pattern.search(inStr)
    if match:
        return match.span()[0]
    else:
        return -1

def reQuote(pattStr):
    "quote special characters in re patter component"
    QUOTE_CHR = '\\'
    l = list(pattStr)
    l = [QUOTE_CHR + c for c in l]
    return ''.join(l)


def addrNameFromString(inStr):
    """parses name and address from string

       formats:
          'name <address>'
          'addrSameAsName'
    """
    l = inStr.split('<')
    if len(l) == 1:
        name = addr = inStr
    elif len(l) == 2:
        name = l[0]
        addr = l[1]
    else:
        raise Exception('unexpected format of addrStr: %s' % inStr)
    name = name.strip()
    if addr[-1] == '>':  
        addr = addr[:-1]
    addr = addr.strip()
    return (addr, name)


#class MappingSpec(list):
#    "how to map syntactic tokens to elements"
#
#    def __init__(self):
#        list.__init__(self)
#        self._startTagPattern = None
#        self._mappingDict = None
#
#    def mappingDict(self):
#        "mappingSpecItems indexed by startTag for easy use"
#        if not self._mappingDict:
#            self._mappingDict = dict([[i.startTag, i] for i in self])
#        return self._mappingDict
#
#    def startTagPattern(self):
#        if not self._startTagPattern:
#            self._startTagPattern = tagsPattern(self.mappingDict.keys())
#        return self._startTagPattern
#
#    def append(self, item):
#        list.append(self, item)
#        self._mappingDict = None
#        self._startTagPattern = None
#
#class MappingSpecItem(object):
#    def __init__(self, startTag, endTag, mappingFunc, singleWord=False):
#        """start and endTag are strings, 
#           mappingFunc maps a txt to a list of nodes
#        """
#        self.startTag = startTag
#        self.endTag = endTag
#        self.func = mappingFunc
#        self.singleWord = singleWord
#        self.endTagPattern = endTagPattern(self.endTag, self.singleWord)
#
#
# 
#
#
#    



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
=========
BaseNodes
=========

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

This module provides basic nodes of different types to build a parse tree.  It
includes some methods to navigate through trees and inspect elements.  

It is agnostic of any Writer capability, i.e., knowledge of how to render nodes
in different output formats (xml, html, other).

Writers and Transformers are meant to subclass the classes of this module to
add writer/transformer-specific methods.  
"""


import pprint

#import re

#== Utility Classes ==================================================

class MultiDict(object):
    def __init__(self, _dict= None):
        _dict = _dict or {}
        self._dict = {}
        for k, v in _dict.items():
            self.insert(k, v)

    def __nonzero__(self):
        return bool(self._dict)

    def asDict(self):
        return self._dict
    
    def __eq__(self, other):
        return self.asDict() == other.asDict()

    def insert(self, key, val):
        if key in self._dict:
            self._dict[key].append(val)
        else:
            self._dict[key] = [val]

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self.insert(key, value)

    def items(self):
        return self._dict.items()

    def update(self, mDict):
        for k, vList in mDict.items():
            for v in vList:
                self.insert(k, v)

    def clone(self):
        "returns a different instance with equal content"
        md = MultiDict()
        md._dict = self._dict.copy()
        return md

    def contains(self, _dict):
        "True if all dict items are included"
        for key, val in _dict.items():
            if not key in self._dict or not val in self._dict[key]:
                return False
        return True

  
#== Node Classes ==================================================

class Node(object):
    def __init__(self, txt=''):
        self.children = []
        self.parent = None
        self.txt = txt

    def __repr__(self):
        if not self.parent and self.__class__.__name__ == 'Node':
            detailStr = "I'm the root"
        elif self.txt:
            if len(self.txt) < 9: 
                detailStr = '"%s"' % self.txt
            else:
                detailStr = '"%s.."' % self.txt[:8]
        else:
            detailStr = ''
        if detailStr: 
            detailStr = ": %s" % detailStr
        return "<%s instance%s>" % (self.__class__.__name__, detailStr)

    def addChild(self, child):
        self.children.append(child)
        child.parent = self

    def delChild(self, child):
        child.parent = None
        self.children.remove(child)

    def _updateParentInChildren(self):
        "updates the parent in all children to self"
        for child in self.children:
            child.parent = self

    def setChildren(self, childL):
        "sets children to given list of children"
        self.children = childL
        self._updateParentInChildren()

    def replaceWith(self, substNodeOrList):
        "replaces self with substNode or a list of Nodes"
        # if node has children, they are attached at the first
        # node in the list
        childNdx = self.parent.children.index(self)
        if isinstance(substNodeOrList, list):
            nodeL = substNodeOrList
        else:
            nodeL = [substNodeOrList]
        for node in nodeL:
            node.parent = self.parent
        if self.children:
            nodeL[0].children = self.children
            nodeL[0]._updateParentInChildren()
        self.parent.children[childNdx:childNdx+1] = nodeL


    def __getitem__(self, key):
        return self.children[key]

    def asNestedList(self):
        if self.children:
            childReps = [c.asNestedList() for c in self.children]
            return [self, childReps]
        else:
            #return [self]
            return self

    def pprint(self, indent=2):
        pprint.pprint(self.asNestedList(), indent=indent)

    def _isEligibleEl(self, tagVals, attVals, elClassName):
        return False 
        #baseNode never shows up
        #specialize in subclass

    def findAllEls(self, tagVals=None, attVals=None, elClassName='Element'):
        """returns a list of all ancester Elements
           optionally of given tag or attribute values
        """
        tagVals = tagVals or []
        attVals = attVals or {}
        l = []
        if self._isEligibleEl(tagVals, attVals, elClassName):
            l.append(self)
        for child in self.children:
            l.extend(child.findAllEls(tagVals, attVals))
        return l

    def _isEligibleNode(self, nodeType, selectFunct):
        if nodeType and not self.__class__.__name__ == nodeType:
            return False
        if selectFunct and not selectFunct(self):
            return False
        return True

    def findAllNodes(self, nodeType='', selectFunct=None):
        """returns a list of all ancester Nodes
           of a given type
           selectFunct takes a node as input and returns
           True if it needs to be included, and False otherwise
        """
        l = []
        if self._isEligibleNode(nodeType, selectFunct):
            l.append(self)
        for child in self.children:
            l.extend(child.findAllNodes(nodeType, selectFunct))
        return l


    def lastNonTextNode(self):
        "decends in a line of single decendents (children[0]) and returns the last non-Text-Node"
        lastN = self
        while lastN.children and not isinstance(lastN.children[0], Text):
            lastN = lastN.children[0]
        return lastN

#---------------------------------------------------------------------
class GenEl(Node):
    "a generic element (not xml-elemnt) that is not used by itself"

    def __init__(self, tag='', txt='', atts=None):
        "att is dict or MultiDict"
        Node.__init__(self, txt)
        if atts:
           if atts.__class__.__name__ == 'dict':
               myAtts = MultiDict(atts)
           else:
               myAtts = atts.clone()
        else:
           myAtts = MultiDict()
        self.tag = tag
        self.atts = myAtts

    def attDict(self):
        return self.atts.asDict()

    def addAttr(self, key, val):
        self.atts[key] = val

    def __repr__(self):
        if self.atts.asDict():
            detailStr = ': %s' % self.attDict()
        else: 
            detailStr = ''
        l = []
        classStr = self.__class__.__name__.split('.')[-1]
        if self.tag: l.append(self.tag)
        l.append(classStr)
        startStr = ' '.join(l)
        return '<%s%s>' % (startStr, detailStr)

    def _isEligibleEl(self, tagVals, attVals, elClassName):
        if self.__class__.__name__ != elClassName:
            return False
        if tagVals and not self.tag in tagVals:
            return False
        if attVals and not self.atts.contains(attVals):
            return False
        return True

#---------------------------------------------------------------------
class Text(Node): 
    pass


#-- Utility Functions ----------------------------

def linkNodes(listOfDecendents):
    "links all elements in parent child relationship"
    nodesRev = listOfDecendents[::-1]
    parent = nodesRev.pop()
    #print "!!! linkNodes: starting with %s" % parent
    while nodesRev:
        el = nodesRev.pop()
        #print "!!! linkNodes: adding %s to %s" % (el, parent)
        parent.addChild(el)
        parent = el



from inlineParser import endTagPattern, tagsPattern


class MappingSpec(list):
    "how to map syntactic tokens to elements"

    def __init__(self):
        list.__init__(self)
        self._startTagPattern = None
        self._mappingDict = None

    def mappingDict(self):
        "mappingSpecItems indexed by startTag for easy use"
        if not self._mappingDict:
            self._mappingDict = dict([[i.startTag, i] for i in self])
        return self._mappingDict

    def startTagPattern(self):
        if not self._startTagPattern:
            self._startTagPattern = tagsPattern(self.mappingDict().keys())
        return self._startTagPattern

    def append(self, item):
        list.append(self, item)
        self._mappingDict = None
        self._startTagPattern = None

class MappingSpecItem(object):
    def __init__(self, startTag, endTag, handler, singleWord=False):
        """start and endTag are strings, 
           mappingFunc maps a txt to a list of nodes
        """
        self.startTag = startTag
        self.endTag = endTag
        self.handler = handler
        self.singleWord = singleWord
        self.endTagPattern = endTagPattern(self.endTag, self.singleWord)





from __future__ import absolute_import
from ..buml.mappingSpec import *


def test_mappingSpec():
    "mappingSpec: mappingSpec and Item"
    def emHandler(): pass
    def strongHandler(): pass
    m = MappingSpec()
    assert m == []
    m.append(MappingSpecItem('*', '*', emHandler))
    assert m._startTagPattern == None
    assert m._mappingDict == None
    assert m.mappingDict().keys() == ['*']
    m.append(MappingSpecItem('**', '**', strongHandler))
    assert m._startTagPattern == None
    assert m._mappingDict == None
    assert m.mappingDict().keys() == ['*', '**']
    assert m.startTagPattern is not None



from docutils.examples import html_body

def rst(rstStr):
    "ReST parser from docutils"
    rendered = html_body(rstStr, input_encoding='utf-8', 
                         output_encoding='utf-8').strip()
    if rendered.startswith('<div class="document">\n'):
        rendered = rendered[23:]
        if rendered.endswith('\n</div>'):
            rendered = rendered[:-7]
    #alternative: find first > and last < and cut off there...
    return rendered



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
=============
bumdProcessor
=============

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

Unittests for buml
"""
from __future__ import absolute_import

import unittest
from ..buml.bumlParser import Parser

#== Tests ========================================

#def testNoseFindsMe():
#    "test-all: Nose Finds Me"
#    assert 0

def test_allowedCharacters():
    "all: tests allowed characters"
    p = Parser()
    l = []
    #-- any char but " in lineText of simpleEl: (bumlComment)
    s = """/ buml comment " \' & > . # @  / \\ !"""
    o = ""
    l.append((s, o))
    #-- start normally commented section----------------------
    #-- pre: single line el > is declared by nodeFactory
    #-- uncomment this only if you uncomment the corresponding
    #   modifications in nodeFactory
    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #-- any char but " as specialChar of simpleEl:
    # s = """> very special singleLineEl " \' & > . # @ """
    # o = """<!-- very special singleLineEl " \' & > . # @ -->"""
    # l.append((s, o))
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    #-- any char but " in lineText of simpleEl: (XmlComment)
    s = """! very special singleLineEl " \' & > . # @ / \\ ! """
    o = """<!-- very special singleLineEl " \' & > . # @ / \\ ! -->"""
    l.append((s, o))
    #--
    #-- start normally commented section----------------------
    #-- pre: disable validity checking in nodeFactory
    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #-- any character but " & > . # @ in elName
    #s = """elName/\!:= blank"""
    #o = """<elName/\\!:= blank/>"""
    #l.append((s, o))
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    #-- any character but " & > . # @ = in attName
    s = """ @start ' / \\ ! : - end"""
    o = """ <div start ' / \\ ! : - end="True"/>"""
    l.append((s, o))
    #-- any character but " & > . # @ in unQuotedAttVal
    s = """ .start ' / \\ ! : - < end"""
    o = """ <div class="start ' / \\ ! : - < end"/>"""
    l.append((s, o))
    #-- any character but " & > ' in quotedAttVal
    s = """@att='start / \\ ! : - < . # @ end'"""
    o = """ <div att="start / \\ ! : - < . # @ end"/>"""
    l.append((s, o))
    #---------------------------------
    for (s, o) in l:
        s = s.replace("'''", '"""')
        o = o.strip()
        out = p.parseBuml(s).xml().strip()
        print "out : %s" % out
        print "o   : %s" % o
        assert out == o

def test_emptyElementMode():
    """all: emptyElementMode"""
    p = Parser()
    rootN = p.parseBuml("br")
    rootN.setEmptyElementMode('sgml')
    out = rootN.xml().strip()
    print out
    assert out == "<br>"
    #--
    rootN.setEmptyElementMode('xml')
    out = rootN.xml().strip()
    print out
    assert out == "<br/>"
    #--
    rootN.setEmptyElementMode('compat')
    out = rootN.xml().strip()
    print out
    assert out == "<br />"



def test_ParserAll():
    """all: parser b2x examples"""
    #--
    p = Parser()
    #-- configuration that is expected in the tests!!!
    dummy = p.parseBuml('br')
    dummy.setEmptyElementMode('xml')
    dummy.setRenderMode('pretty', 2)
    #--
    l = []
    s = "/a buml comment"
    x = ""
    l.append((s, x))
    s = "! an html comment"
    x = '<!-- an html comment -->'
    l.append((s, x))
    s = """
html
  body
    h1 "Welcome to buml
    p '''This is text
         that spans
         multiple lines""" 
    x = """
<html>
  <body>
    <h1>Welcome to buml</h1>
    <p>
      This is text
      that spans
      multiple lines
    </p>
  </body>
</html>"""
    l.append((s, x))
    s = 'p.cool "This is a cool paragraph\n' + \
        'p.cool.blue "this has two classes'
    x = '<p class="cool">This is a cool paragraph</p>\n' + \
        '<p class="cool blue">this has two classes</p>'
    l.append((s, x))
    s = 'div#navbar "bar here'
    x = '<div id="navbar">bar here</div>'
    l.append((s, x))
    s = "\n".join([
        """h1@myAttr=something "title 1""", 
        """h2@att2='two words' "title 2""", 
        """img@src='me.png'""", 
        """a@href='else#where' "go""", 
        """h3@flag "set to True""",
        """h4@a1=v1@a2=v2 "two attributes""" ])
    x =  "\n".join([
        '<h1 myAttr="something">title 1</h1>',
        '<h2 att2="two words">title 2</h2>',
        '<img src="me.png"/>',
        '<a href="else#where">go</a>',
        '<h3 flag="True">set to True</h3>',
        '<h4 a1="v1" a2="v2">two attributes</h4>' ])
    l.append((s, x))
    s = '.class1 "A dev element\n#id1 "this too'
    x = '<div class="class1">A dev element</div>\n' + \
        '<div id="id1">this too</div>'
    l.append((s, x))
    s = '.cl1#id1@att1=v1.cl2 "all mixed'
    x = '<div class="cl1 cl2" att1="v1" id="id1">all mixed</div>'
    l.append((s, x))
    s = 'br\nhr'
    x = '<br/>\n<hr/>'
    l.append((s, x))
    s = """ul > li > a@href='this.html' "first item\n""" + \
          '  li "second item'
    x = '<ul>\n  <li>\n    <a href="this.html">first item</a>\n' + \
        '  </li>\n  <li>second item</li>\n</ul>'
    l.append((s, x))
    s = 'br & br & hr & br'
    x = '<br/>\n<br/>\n<hr/>\n<br/>'
    l.append((s, x))
    s = """div > img@src='me.png' & br & br & ul > li "only one item"""
    x = '<div>\n  <img src="me.png"/>\n</div>\n' + \
        '<br/>\n<br/>\n<ul>\n  <li>only one item</li>\n</ul>'
    l.append((s, x))
    s = 'p\n' + \
        '  "this is a first line\n' + \
        '  em "something emphasized\n' + \
        '  " and\n' + \
        '  strong "something strong\n' + \
        '  "and some tail'
    x = '<p>\n' + \
        '  this is a first line\n' + \
        '  <em>something emphasized</em>\n' + \
        '   and\n' + \
        '  <strong>something strong</strong>\n' + \
        '  and some tail\n' + \
        '</p>'
    l.append((s, x))

    s = """
.cooltext '''asdf 0708 ~!@#$%^&*()_+`\|/?,./<>-=_+
             and even " and ''' without closing anything"""
    x = """
<div class="cooltext">
  asdf 0708 ~!@#$%^&*()_+`\|/?,./<>-=_+
  and even " and ''' without closing anything
</div> """
    s = s.replace("'''", '"""')
    x = x.replace("'''", '"""')
    l.append((s, x))
    s = """
.indent '''and empty lines:
         
           and any kind of indentation
             relative 
               to 
                 the
              start
                of 
           the 
                      first
             line"""
    x = """
<div class="indent">
  and empty lines:
  
  and any kind of indentation
    relative 
      to 
        the
     start
       of 
  the 
             first
    line
</div> """
    s = s.replace("'''", '"""')
    l.append((s, x))
    s = """
div '''<em>evidently</em> you can
       embedd <html></html> like
       <a href="http://bruegger.it">visit</a>"""
    x = """
<div>
  <em>evidently</em> you can
  embedd <html></html> like
  <a href="http://bruegger.it">visit</a>
</div>"""
    s = s.replace("'''", '"""')
    l.append((s, x))

    s = """
div '''and you can use *inline* **markup**
       like `reStructured Text 
       <http://docutils.sourceforge.net/rst.htm>`_

       ..topic:: Interesting
   
          That you can pass to some
          configurable **inline parser**"""
    x = """
<div>
  and you can use *inline* **markup**
  like `reStructured Text 
  <http://docutils.sourceforge.net/rst.htm>`_
  
  ..topic:: Interesting
  
     That you can pass to some
     configurable **inline parser**
</div>"""
    s = s.replace("'''", '"""')
    l.append((s, x))
    s = """
{% these are treated
   {{ just like normal elements"""
    x = """
{% these are treated %}
  {{ just like normal elements }}"""
    l.append((s, x))

    s = """
        """
    x = """
        """
    s = s.replace("'''", '"""')
    l.append((s, x))

    s = """
        """
    x = """
        """
    s = s.replace("'''", '"""')
    l.append((s, x))
    s = ''
    x = ''
    l.append((s, x))
    s = ''
    x = ''
    l.append((s, x))
    #----------
    for (s, x) in l:
        s = s.replace("'''", '"""')
        x = x.strip()
        out = p.parseBuml(s).xml().strip()
        print "=========="
        print out
        print "------"
        print x
        assert out == x



from __future__ import absolute_import
from ..buml.bumlParser import *


#def testNoseFindsMe():
#    "bumlParser: testNoseFindsMe"
#    assert 0


def testTabs():                    
    """bumlParser: tab substitution"""
    l=['\t\txxx']
    assert eliminateTabs(l) == ['        xxx']
    l=['  \txxx']
    assert eliminateTabs(l) == ['      xxx']
    l=['\txxx\t']
    assert eliminateTabs(l) == ['    xxx\t']
    l="""  xx\n  \t  yy\n\n  zz\t\t""".split('\n')
    r="""  xx\n        yy\n\n  zz\t\t""".split('\n')
    assert eliminateTabs(l) == r

def testIndent():                    
    """bumlParser: indentParsing"""
    s="   xx yy zz "
    assert indent(s) == 3

def testSubRanges():                    
    """bumlParser: subLineRange and subRanges"""
    #empty lines:  not possible, eliminated as pre-condition
    l = ["hello", "  there", "     how", "  are", "     you ", 
         "       doing", "Bud"]
    mono = ["hello", "  there", "     how", "  are", "     you ", 
            "       doing"]
    rest = ['Bud']
    assert subLineRange(l) == (mono, rest)
    l = ["hello"]
    assert subLineRange(l) == (['hello'], [])
    l = ["  hello", "    one", "    two", "  there", 
         "  buddy", "    over", "      there"]
    r1 = ["  hello", "    one", "    two"]
    r2 = ["  there"]
    r3 = ["  buddy", "    over", "      there"]
    assert subRanges(l) == ([r1, r2, r3])
    l = ["  hello", "", "    one", "", "  there"]
    r1 = ["  hello", "", "    one", ""]
    r2 = ["  there"]
    assert subRanges(l) == ([r1, r2])


def testTagTextSplit():                    
    """bumlParser: tab text splitting"""
    p=Parser()
    s="""  div.cool@title='two words' otherTag#id1 "text here"""
    tag="div.cool@title='two words' otherTag#id1"
    txt='text here'
    assert p._tagTextSplit(s) == (tag, txt, False, 43)
    s="""  div.cool  otherTag#id1"""
    tag="""div.cool  otherTag#id1"""
    txt=''
    assert p._tagTextSplit(s) == (tag, txt, False, -1)
    s='  div.cool  otherTag#id1 """start of multiline text'
    tag='div.cool  otherTag#id1'
    txt='start of multiline text'
    assert p._tagTextSplit(s) == (tag, txt, True, 28)
    s='    """only start of multiline text'
    tag=''
    txt='only start of multiline text'
    assert p._tagTextSplit(s) == (tag, txt, True, 7)
    s='   "only start of text  '
    tag=''
    txt='only start of text  '
    assert p._tagTextSplit(s) == (tag, txt, False, 4)

def testTagRestSplit():
    """bumlParser: tagRestSplit"""
    p=Parser()
    s="span.blue"
    assert p._tagRestSplit(s) == ('span', '.blue')
    s="div"
    assert p._tagRestSplit(s) == ('div', '')
    s=".blue"
    assert p._tagRestSplit(s) == ('div', '.blue')

def testTagAttrSplit():
    """bumlParser: tagAttrSplit"""
    p=Parser()
    s=".cool#id1@title='x'"
    assert p._tagAttrSplit(s) == ('class', 'cool', "#id1@title='x'")
    s="#id1@title='x'"
    assert p._tagAttrSplit(s) == ('id', 'id1', "@title='x'")
    s="@title='x'.hello"
    assert p._tagAttrSplit(s) == ('title', 'x', ".hello")
    s="@title='x'"
    assert p._tagAttrSplit(s) == ('title', 'x', "")
    s="@src='img.png'.cl1"
    assert p._tagAttrSplit(s) == ('src', 'img.png', ".cl1")
    s="@href='else#where'#id1"
    assert p._tagAttrSplit(s) == ('href', 'else#where', "#id1")
    s="@mail='user@gmail.com'@other=xy"
    assert p._tagAttrSplit(s) == ('mail', 'user@gmail.com', "@other=xy")

def testTagParse():
    """bumlParser: tagParse"""
    p=Parser()
    s='.hello'
    el = p._tagParse(s)
    assert el.tag == 'div'
    assert el.attDict()['class'] == ['hello']
    s='span#id1'
    el = p._tagParse(s)
    assert el.tag == 'span'
    assert el.attDict()['id'] == ['id1']

def testTagsParse():
    """tagsParse"""
    p = Parser()
    s='.hello > .blue > ul.big>li#id3'
    els = p._tagsParse(s)
    assert len(els) == 1
    assert els[0].tag == 'div'
    assert els[0][0].tag == 'div'
    assert els[0][0][0].tag == 'ul'
    assert els[0][0][0][0].tag == 'li'
    assert els[0][0][0][0].attDict() == {'id':['id3']}
    s='.hello > .blue & br&ul>li'
    els = p._tagsParse(s)
    assert els[0].tag == 'div'
    assert els[1].tag == 'br'
    assert els[2].tag == 'ul'
    assert els[2][0].tag == 'li'
    assert els[2][0].children == []

def testGetAtts():
    """bumlParser: getAtts"""
    p=Parser()
    s=".cool#id1@title='x'"
    md = MultiDict({'class' : 'cool', 'id': 'id1', 'title': 'x'})
    assert p._getAtts(s) == md
    s=".cool#id1.cooler.coolest"
    md = MultiDict({'class' : 'cool', 'id': 'id1'})
    md.insert("class", 'cooler')
    md.insert("class", 'coolest')
    assert p._getAtts(s) == md

def testParseGenAttr():
    """bumlParser: parseGenAttr"""
    p = Parser()
    s="title=hello"
    assert p._parseGenAttr(s) == ('title', 'hello')
    s="words='one and two'"
    assert p._parseGenAttr(s) == ('words', 'one and two')
    s="flag"
    assert p._parseGenAttr(s) == ('flag', 'True')

def testTagDecendingSplit():
    """bumlParser: tagDecendingSplit"""
    p = Parser()
    s="div.cool > #id1> table>tbody.xx>trow#n1"
    l = ["div.cool", "#id1", "table", "tbody.xx", "trow#n1"]
    assert p._tagDecendingSplit(s) == l

def testTagSameLevelSplit():
    """bumlParser: tagSameLevelSplit"""
    p = Parser()
    s="div.cool > #id1 & br & ul>li.odd"
    l = ["div.cool > #id1", "br", "ul>li.odd"]
    assert p._tagSameLevelSplit(s) == l

def testParseMultiLineTxt():
    """bumlParser: parseMultiLineTxt """
    p = Parser()
    l = ['       =text starts here  ',
         '        and goes over multiple\t',
         '   ',
         '        lines that are lined up  ',
         '          or  ',
         '            indent',
         '              as they want']
    startPos = 8
    firstLine = l[0][startPos:]
    txt = parseMultiLineTxt(firstLine, startPos, l[1:])
    x = ['text starts here  ',
         'and goes over multiple\t',
         '',
         'lines that are lined up  ',
         '  or  ',
         '    indent',
         '      as they want']
    t = '\n'.join(x)
    assert txt == t

def test_nodeFactoryPassing():
    "bumlParser: nodeFactoryPassing"
    p = Parser()
    assert p.nodeFactory.__class__.__name__ == "NodeFactory"
    class FakeNodeFactory:
        pass
    fnf = FakeNodeFactory()
    pp = Parser(fnf)
    print pp.nodeFactory.__class__.__name__ 
    assert pp.nodeFactory.__class__.__name__ == "FakeNodeFactory"

def test_updateAfterReConfig():
    "bumlParser: updateAfterReConfig"
    p = Parser()
    patt1 = p.TAG_PART_SEPARATORS.pattern
    p._SAME_LEV_SEP = '+'
    p.updateAfterReConfig()
    patt2 = p.TAG_PART_SEPARATORS.pattern
    assert patt1 == patt2 #_SAME_LEV_SEP not part of pattern
    p._ID_CHR = '$'
    p.updateAfterReConfig()
    patt3 = p.TAG_PART_SEPARATORS.pattern
    print patt1
    print patt3
    assert patt1 == '[\#\.\@]'
    assert patt3 == '[\$\.\@]'


    



from __future__ import absolute_import
from ..buml.inlineParser import *
from ..buml.inlineParser import _tagRestSplit
from ..buml.inlineMapper import InlineMapper
from ..buml.transformUtils import applyInlineMarkup

#def testNoseFindsMe():
#    "inlineParser: testNoseFindsMe"
#    assert 0

def test_orderedSepStrs():
    "inlineParser: orderedSepStrs"
    l=['*', ':::', '***', ':', '""', '**']
    print orderedSepStrs(l)
    assert orderedSepStrs(l) == [':::', ':', '***', '**', '*', '""']

def test_tagsPattern():
    "inlineParser: tagsPattern"
    p = tagsPattern(['a', 'b', 'c'])
    print p.pattern
    assert p.pattern == '^(.*?)(\\c|\\b|\\a)'
    p = tagsPattern(['b', 'a', 'c'])
    print p.pattern
    assert p.pattern == '^(.*?)(\\c|\\b|\\a)'

def test_endTagPattern():
    "inlineParser: endTagPattern"
    p = endTagPattern(']]', singleWord=True)
    expected = '^(\\S*?)(\]\])'
    print "computed: %s" % p.pattern
    print "expected: %s" % expected
    assert p.pattern == expected
    #--
    p = endTagPattern(']]', singleWord=False)
    expected = '^(.*?)(\]\])'
    print "computed: %s" % p.pattern
    print "expected: %s" % expected
    assert p.pattern == expected

def _findTagAssertion(inStr, patt, expected):
    print ">>%s<<" % inStr
    print patt.pattern
    res = findTag(patt, inStr)
    print "findTag:  pos=%s, txt='%s', tag='%s'" % res
    print "expected: pos=%s, txt='%s', tag='%s'" % expected
    print '--'
    assert res == expected

def test_findEndTag():
    "inlineParser: findStartTag"
    #--------------------------------
    #-- a single word
    patt = endTagPattern(']]', True)
    #--
    lineStr = 'single]] rest'
    expected = (8, 'single', ']]')
    _findTagAssertion(lineStr, patt, expected)
    #-
    lineStr = 'two words]] rest'
    expected = (-1, lineStr, '')
    _findTagAssertion(lineStr, patt, expected)
    #-
    lineStr = ']] rest'
    expected = (2, '', ']]')
    _findTagAssertion(lineStr, patt, expected)
    #--------------------------------
    #-- multiple words
    patt = endTagPattern(']]', False)
    #--
    lineStr = 'two words]] rest'
    expected = (11, 'two words', ']]')
    _findTagAssertion(lineStr, patt, expected)
    #--

def test_findStartTag():
    "inlineParser: findStartTag"
    tagL = ['**', '***', '::']
    patt = tagsPattern(tagL)
    #--
    lineStr = 'ads ***ital*** and **bold** and *ital* and ::repl:: cool'
    expected = (7, 'ads ', '***')
    _findTagAssertion(lineStr, patt, expected)
    #--
    tagL = ['::', '**', '***']
    patt = tagsPattern(tagL)
    lineStr = 'ads ***ital*** and **bold** and *ital* and ::repl:: cool'
    expected = (7, 'ads ', '***')
    _findTagAssertion(lineStr, patt, expected)
    #--
    lineStr = 'ads *ital* and **bold** and *ital* and ::repl:: cool'
    expected = (17, 'ads *ital* and ', '**')
    _findTagAssertion(lineStr, patt, expected)
    #--
    lineStr = '::ads ***ital*** and'
    expected = (2, '', '::')
    _findTagAssertion(lineStr, patt, expected)
    #--
    lineStr = 'nothing to be found here!'
    expected = (-1, lineStr, '')
    _findTagAssertion(lineStr, patt, expected)
    #--
    lineStr = 'this is on\nsecond **line**!'
    print 
    expected = (20, 'this is on\nsecond ', '**')
    _findTagAssertion(lineStr, patt, expected)

def test_tagRestSplit():
    "inlineParser: _tagRestSplit"
    spec = InlineMapper().mappingSpec()
    print spec.mappingDict().keys()
    s = 'leadingTxt **important** rest'
    print _tagRestSplit(s, spec)
    res = _tagRestSplit(s, spec)
    assert res == ('leadingTxt ', '**', 'important', ' rest')
    #--
    s = '**important** rest'
    print _tagRestSplit(s, spec)
    res = _tagRestSplit(s, spec)
    assert res == ('', '**', 'important', ' rest')
    #--
    s = 'leading **important]] rest'
    print _tagRestSplit(s, spec)
    res = _tagRestSplit(s, spec)
    assert res == ('leading **important]] rest', '', '', '')
    #--
    s = 'nothing to be found'
    print _tagRestSplit(s, spec)
    res = _tagRestSplit(s, spec)
    assert res == ('nothing to be found', '', '', '')

def test_parser():
    "inlineParser: parser"
    spec = InlineMapper().mappingSpec()
    s = 'hello *there* **bold** [[guy <http://guy.org]] over there'
    nodeL = parse(s, spec)
    for node in nodeL:
        node.pprint()
        print
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'hello '
    assert nodeL[1].tag == 'em'
    assert nodeL[1][0].__class__.__name__ == 'Text'
    assert nodeL[1][0].txt == 'there'
    assert nodeL[2].__class__.__name__ == 'Text'
    assert nodeL[2].txt == ' '
    assert nodeL[3].tag == 'strong'
    assert nodeL[3][0].__class__.__name__ == 'Text'
    assert nodeL[3][0].txt == 'bold'
    assert nodeL[4].__class__.__name__ == 'Text'
    assert nodeL[4].txt == ' '
    assert nodeL[5].tag == 'a'
    assert nodeL[5].atts['href'] == ['http://guy.org']
    assert nodeL[5][0].__class__.__name__ == 'Text'
    assert nodeL[5][0].txt == 'guy'
    assert nodeL[6].__class__.__name__ == 'Text'
    assert nodeL[6].txt == ' over there'
    #--
    s = 'please, ::replace:: this'
    nodeL = parse(s, spec)
    print '--------'
    for node in nodeL:
        node.pprint()
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'please, '
    assert nodeL[1].tag == 'replace'
    assert nodeL[1][0].__class__.__name__ == 'Text'
    assert nodeL[1][0].txt == 'replace'
    #-- 
    s = 'non * matching endTag ::match:: end'
    nodeL = parse(s, spec)
    print ">> nodeL: ", nodeL
    print '--------'
    for node in nodeL:
        node.pprint()
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'non * matching endTag '
    assert nodeL[1].tag == 'replace'
    s = 'non * matching endTag and no more tags'
    nodeL = parse(s, spec)
    print '--------'
    for node in nodeL:
        node.pprint()
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'non * matching endTag and no more tags'

def test_parserRecursion():
    "inlineParser: parser recursive decend"
    spec = InlineMapper().mappingSpec()
    s = 'hello ** bold *buml ::replMe:: markup* good-bye**'
    nodeL = parse(s, spec)
    for node in nodeL:
        node.pprint()
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'hello '
    assert nodeL[1].__class__.__name__ == 'Element'
    assert nodeL[1].tag == 'strong'
    assert nodeL[1][0].__class__.__name__ == 'Text'
    assert nodeL[1][0].txt == ' bold '
    assert nodeL[1][2].__class__.__name__ == 'Text'
    assert nodeL[1][2].txt == ' good-bye'
    assert nodeL[1][1].__class__.__name__ == 'Element'
    assert nodeL[1][1].tag == 'em'
    assert nodeL[1][1][0].__class__.__name__ == 'Text'
    assert nodeL[1][1][0].txt == 'buml '
    assert nodeL[1][1][2].__class__.__name__ == 'Text'
    assert nodeL[1][1][2].txt == ' markup'
    assert nodeL[1][1][1].__class__.__name__ == 'Element'
    assert nodeL[1][1][1].tag == 'replace'
    assert nodeL[1][1][1][0].__class__.__name__ == 'Text'
    assert nodeL[1][1][1][0].txt == 'replMe'


def test_applyInlineMarkup():
    "inlineParser: transformUtils.applyInlineMarkup"
    im = InlineMapper()
    spec= im.mappingSpec()
    nf = im.nodeFactory
    el = nf.mkElementNode('div')
    s = 'hello ** bold *buml ::replMe:: markup* good-bye**'
    txtN = nf.mkTextNode(s)
    el.addChild(txtN)
    #-- 
    print "before:"
    print el.pprint()
    applyInlineMarkup(el, spec)
    print "after:"
    print el.pprint()
    assert el.__class__.__name__ == 'Element'
    assert el.tag == 'div'
    nodeL = el.children #so I can copy assertions from above
    assert nodeL[0].__class__.__name__ == 'Text'
    assert nodeL[0].txt == 'hello '
    assert nodeL[1].__class__.__name__ == 'Element'
    assert nodeL[1].tag == 'strong'
    assert nodeL[1][0].__class__.__name__ == 'Text'
    assert nodeL[1][0].txt == ' bold '
    assert nodeL[1][2].__class__.__name__ == 'Text'
    assert nodeL[1][2].txt == ' good-bye'
    assert nodeL[1][1].__class__.__name__ == 'Element'
    assert nodeL[1][1].tag == 'em'
    assert nodeL[1][1][0].__class__.__name__ == 'Text'
    assert nodeL[1][1][0].txt == 'buml '
    assert nodeL[1][1][2].__class__.__name__ == 'Text'
    assert nodeL[1][1][2].txt == ' markup'
    assert nodeL[1][1][1].__class__.__name__ == 'Element'
    assert nodeL[1][1][1].tag == 'replace'
    assert nodeL[1][1][1][0].__class__.__name__ == 'Text'
    assert nodeL[1][1][1][0].txt == 'replMe'




#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
============
inlineParser
============

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

This is the standard inline parser that comes with buml.

What is actually done by the parser depends on a mappingSpec as typically
provided by the mappingSpec method of the inlineMapper.InlineMapper.  

The parser can be applied to a tree of nodes using
tranformUtils.applyInlineMarkup.
"""

#  ToDo:
#  ----
#  
#  * support for quoting special characters with \

import re
from utilities import reQuote

#============================================================

def endTagPattern(endTag, singleWord):
    """
    matches any text (non-greedy) from start to endTag.
    Single word restricts match to a single word
    """
    if singleWord: 
        txtChr = '\\S'
    else:
        txtChr = '.'
    patternStr = '^(%s*?)(%s)' % (txtChr, reQuote(endTag))
    return re.compile(patternStr, re.DOTALL)

def orderedSepStrs(sepStrL):
    "orderes such that higher repetitions of the same char come first"
    # ['*', '**'] is bad,  ['**', '*'] works
    # this is necessary to make the re match "greedy"
    l = sepStrL[:]  #a copy
    if '' in l:
        l.remove('')
    l.sort()
    l.reverse()
    return l

def tagsPattern(tags):
    "returns a compiled re pattern to match the first tag"
    tags = orderedSepStrs(tags)
    return reStartTagPattern(tags)


def reStartTagPattern(tagL):
    "returns an re pattern that matches any string in the list"
    tagL = [reQuote(s) for s in tagL]
    patternStr = '^(.*?)(%s)' % '|'.join(tagL)
    patt = re.compile(patternStr, re.DOTALL)
    return patt


def findTag(tagPattern, lineStr):
    """
     returns the end position of all matched text (or -1 if no match), 
     the leadingTxt, and matched tag"
    """
    match = tagPattern.match(lineStr)
    if match is not None:
        txt, tag = match.groups()
        return (match.end(), txt, tag)
    else:
        return (-1, lineStr, '')
    
def _tagRestSplit(txtStr, mappingSpec):
    "splits of first (leadingTxt, tag, tagContent, rest) from rest of txtStr"
    startTagEndPos, leadingTxt, tag = findTag(mappingSpec.startTagPattern(), txtStr)
    if startTagEndPos == -1:
        return (txtStr, '', '', '')
    else:
        restStr = txtStr[startTagEndPos:]
        endTagEndPos, content, endTag = findTag(mappingSpec.mappingDict()[tag].endTagPattern, \
                                          restStr)
        if endTagEndPos == -1:
            leadingTxtOfRest, tag, content, restStr = _tagRestSplit(restStr, mappingSpec)
            return (txtStr[:startTagEndPos] + leadingTxtOfRest, tag, content, restStr)
        else:
            return (leadingTxt, tag, content, restStr[endTagEndPos:])

def parse(txt, mappingSpec):
    "parses txt and returns ElementNodes"
    specDict = mappingSpec.mappingDict()
    rest = txt
    nodeL = []
    while rest:
        leadingTxt, tag, content, rest = _tagRestSplit(rest, mappingSpec)
        if leadingTxt:
            nodeL.extend(specDict[''].handler(leadingTxt))
        if tag:
            nodeL.extend(specDict[tag].handler(content))
    return nodeL



from setuptools import setup, find_packages
import sys, os

version = '0.1.4'

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='buml',
      version=version,
      description="buml markup language",
      long_description=read('README'),
      classifiers=['Development Status :: 4 - Beta', 
                   'Environment :: Console',
                   'Environment :: Web Environment', 
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2.7',
                   'Topic :: Text Processing :: Markup',
                   'Topic :: Text Processing :: Markup :: XML',
                   'Topic :: Text Processing :: Markup :: HTML',
                   'Topic :: Text Processing' 
                   ],
      keywords='buml xml html nodeTree shpaml template',
      author='Bud P. Bruegger',
      author_email='bud@bruegger.it',
      url='http://bruegger.it/buml/',
      license='FreeBSD',
      packages=['buml', 'tests'],
      scripts=['bin/b2x', 'bin/bi2x'],
      data_files=['testData/demo.buml', 'testData/inlineDemo.buml'],
      include_package_data=True,
      zip_safe=True
      )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

"""
==========
bumlParser
==========

This module is a part of *buml*.

:Author and Copyright:  Bud P. Bruegger
:License:               FreeBSD

The buml parser takes a file in buml syntax and converts it in a tree of nodes
that is represented by the root node that the parser returns.  

The module that defines the node classes needs to be explicitly passed to the
parser function.  Usually, this module is a specialization of baseNodes. While
baseNodes provides the minimal functionality needed by the buml parser but
lacks the functionality of writers (which produce text (html, xml, ..) output
from a parse tree or transformer (which manipulates parse trees).  
"""
 

#  buml -- Bud's markup language
#  -----------------------------
#
#  Inspired by Sphaml (shpaml.webfactional.com) and 
#  Compact XML
#  http://packages.python.org/compactxml/
#
#  Limitations:
#  * '.' in tag name not possible, difficult to fix
#  
#  Every line is ether an element node or a text node.
#  
#  generic element names:          start with any character but  " & > . # @
#                                  ommission of tag name defaults to "div"
#  multiple decending elements:    separated by >
#  multiple same-level elements:   separated by &
#  classAttributes:                start with a .
#  idAttributes:                   start with a #
#  otherAttributes:                start with @ have an att value after =
#                                  multi-word values must be enclosed in '
#                                  omission of = and value defaults to
#                                  =True
#  single line text:               starts with "
#  multi line text:                starts with """
#
# -- the above markup is "built-in" and hard to modify
#
# -- the markup below is and assignment made by nodeFactory and is easy to modify
#                                  (eventually, this may become pluggable)
#
#  comments:                       starts with / on their own lines (also /* //)
#  xmlComments:                    start with !
#  templatingNodes:                start with { and require a second character
#                                  that indicates type: {{ or {%
#                                  (see e.g., http://jinja.pocoo.org/)
#  processingInstr:                start with ?
#  docTypeDecl:                    starts with |
#  CDATA sections:                 only a [ on a line by itself with text content
#
#  ----- not done yet -----
#  lineContinuations: start with \
#
#
#===============================================================
#  Notes on possible Evolution:
#  -------------------
#
#  * need to check for unexpected indent:  
#    - increasing indent steps at least by 2 chars
#    - decreasing lines use previously established indents
#  
#  * the parser should have an argument to pass in an inline parser that
#  processes all text.  As a result, it returns a list of nodes that can be
#  inserted in the tree.  An optional attribute (class is easy to write) may
#  indicate where to apply the inline parser.  
#  NO: THIS IS BETTER DONE BY A TRANSFORMER
#  
#  * idea for inline syntax: ..this is <<el1.class1@att=hi>el2#id1 |
#  http://cnn.com, tool tip | anchor text>> ..inline goes on Very easy to
#  parse: no nesting, simple splits do most of the job, rest already written
#  
#  A sorthand for often used things:  *italic* **bold**  (like in rst).  The
#  parser could have an argument to specify a mapping dictionary that mars
#  the markup ('*', '**', '|') to an element
#  
#  * a writer to ElementTree would interface to other tools
#  
#  * a reader from xml would include (documents or snipplets from other worlds)
#  
#  * maybe maybe a writer that returns an ReST (or Sphinx) parse tree???  But
#  how do I use this?
#  
#  * Sphinx integration:  Embed pages written in buml:
#    - Makefile addition
#    - use role raw
#    - convert buml nodes into ReST nodes ;-)  <== the winner idea
#    - how to use references and index entries??
#  
#  DONE
#  * Use % to indicate Django/Jinja nodes (e.g.,  {% extends "base.html" %},
#    as in http://shpaml.webfactional.com/long_example)
#
#===============================================================

import re
from baseNodes import MultiDict, linkNodes
from utilities import warning, findPatternPos, reQuote
#from utilities import warning, my_import

#== Classes ==================================================

class ParsingError(Exception):
    pass

class UnexpectedError(Exception):
    pass

class ConfigurationError(Exception):
    pass

#== Configuration Section ==================================================
# use configuration functions "config*" and "set*" to change these settings!

#== Utilities =======================================

def eliminateTabsSingleLine(lineStr):
    rest = lineStr.lstrip()
    indentStr = lineStr[:(len(lineStr) - len(rest))]
    if '\t' in indentStr:
        warning("Tab(s) found in ident string: %s" % lineStr)
    indentBlanks = indentStr.replace('\t', '    ')
    return indentBlanks + rest

def eliminateTabs(lines):
    return [eliminateTabsSingleLine(l) for l in lines]

def indent(lineStr):
    return len(lineStr) - len(lineStr.lstrip())

#def findSepPos(str):
#    "returns pos of first separator or -1"
#    match = TAG_PART_SEPARATORS.search(str)
#    if match:
#        return match.span()[0]
#    else:
#        return -1

def isEmpty(someStr):
    return not bool(someStr.strip())

def allLinesEmpty(lines):
    l = [l for l in lines if l.strip()]
    return not bool(l)

#-- parser stuff -------------------------------

def parseMultiLineTxt(txtFirstLine, txtStartPos, otherLines):
    """first line is already without indent, otherLines are raw

       pre: no mixed content
       pre: no empty lines
    """
    linesWOindent = [l[txtStartPos:] for l in otherLines]
    strippedLines = [txtFirstLine]
    strippedLines.extend(linesWOindent)
    return '\n'.join(strippedLines)

def subLineRange(lines):
    """selects the first line and all consecutive lines with
       a lesser indent (all markup corresponding to one node)
       returns selected lines and all non-selected ones as restLines

       note: skips over empty lines
    """
    lenlines = len(lines)
    if len(lines) == 1:
        return (lines, [])
    else:
        firstLineIndent = indent(lines[0])
        ndx = 1
        while ndx < lenlines and ((indent(lines[ndx]) > firstLineIndent) or isEmpty(lines[ndx])):
            ndx += 1
    return (lines[:ndx], lines[ndx:])

def subRanges(lines):
    "gets a list of all ranges at same indent level from lines"
    if len(lines)>0:
        l = []
        lineRange, restLines = subLineRange(lines)
        l.append(lineRange)
        if restLines:
            l.extend(subRanges(restLines))
        return l
    else:
        return []

#== Parser =======================================
class Parser(object):

    def _sepCharPattern(self):
        "returns re pattern for separator characters"
        sepCharStr = self._ID_CHR + self._CLS_CHR + self._ATT_CHR
        sepCharStr = reQuote(sepCharStr)
        pattern = '[%s]' % sepCharStr
        return re.compile(pattern)

    def __init__(self, nodeFactoryInstance=None):
        #module = my_import(nodeFactoryModuleName)
        if nodeFactoryInstance:
            nf = nodeFactoryInstance
        else:
            import nodeFactory
            nf = nodeFactory.NodeFactory()
        self.nodeFactory = nf
        self._TXT_CHR = '"'
        self._SAME_LEV_SEP = '&'  #a string
        self._DECENDING_SEP = '>' #a string
        self._CLS_CHR = '.'
        self._ID_CHR = '#'
        self._ATT_CHR = '@'
        self._EQ_CHR = '='
        self._ATT_QUOTE_CHR = "'"
        #-- update if s.th. changes
        self.TAG_PART_SEPARATORS = self._sepCharPattern()

    def updateAfterReConfig(self):
        """
        reconfigure after changing one or more special characters
        """
        self.TAG_PART_SEPARATORS = self._sepCharPattern()

    def _textParse(self, rawTxtPart, doubleQuotePos):
        "parses text after first double quote"
        if rawTxtPart[:2] == (2 * self._TXT_CHR): #a multi line txt
            isMultiLine = True
            txtPart = rawTxtPart[2:]
            startPos = doubleQuotePos + 3
        else:
            isMultiLine = False
            txtPart = rawTxtPart
            startPos = doubleQuotePos + 1
        return (txtPart, isMultiLine, startPos)

    def _tagTextSplit(self, lineStr):
        """splits tag part from text part
           operates on a full line that includes the indent
    
           returns: (tagSpec, text, isMultiLineTxt, txtStartPos)
        """
        pos = lineStr.find(self._TXT_CHR)
        if pos == -1:  
            return (lineStr.strip(), '', False, pos)
        tagPart = lineStr[:pos].strip()
        txtPart, isMultiLine, startPos = self._textParse(lineStr[pos+1:], pos)
        return (tagPart, txtPart, isMultiLine, startPos)

    def _tagSameLevelSplit(self, tagsStr):
        """same-level tags are separated by &"""
        l = tagsStr.split(self._SAME_LEV_SEP)
        return [i.strip() for i in l]

    def _tagDecendingSplit(self, tagsStr):
        """decending tags are separated by >"""
        l = tagsStr.split(self._DECENDING_SEP)
        return [i.strip() for i in l]
    
    def _parseGenAttr(self, attDef):
        "parses a general attribute"
        l = attDef.split(self._EQ_CHR)
        key = l[0]
        val = l[1] if len(l) == 2 else 'True'
        if val[0] == self._ATT_QUOTE_CHR:
            val = val[1:-1]
        return (key, val)
    
    def _tagAttrSplit(self, attsStr):
        """splits off the first attribute
           after tagName has already been removed
           pre: attsStr not empty
        """
        sepChr = attsStr[0]
        attsStr = attsStr[1:]
        nextSepPos = findPatternPos(attsStr, self.TAG_PART_SEPARATORS)
        #taking care of single-quoted att val
        eqPos = attsStr.find(self._EQ_CHR + self._ATT_QUOTE_CHR) #never mind unquoted att
        if nextSepPos == -1: #no other attribute
            attDef = attsStr
            rest = ''
        elif sepChr == self._ATT_CHR and eqPos < nextSepPos and eqPos != -1:
            endValPos = attsStr.find(self._ATT_QUOTE_CHR, eqPos + 2)
            attDef = attsStr[:endValPos+1]
            rest = attsStr[endValPos+1:]
        else:
            attDef = attsStr[:nextSepPos]
            rest = attsStr[nextSepPos:]
        attVal = attDef
        if sepChr == self._CLS_CHR:
            attKey = 'class'
        elif sepChr == self._ID_CHR:
            attKey = 'id'
        elif sepChr == self._ATT_CHR:
            attKey, attVal = self._parseGenAttr(attDef)
        else:
            raise(UnexpectedError('something went wrong'))
        return (attKey, attVal, rest)


    def _tagRestSplit(self, tagStr):
        "splits off tagName from tagStr"
        match = self.TAG_PART_SEPARATORS.search(tagStr)
        if match:
            pos = match.span()[0]
            tag = tagStr[:pos]
            rest = tagStr[pos:]
        else:
            pos = -1
            tag = tagStr
            rest = ''
        if not tag: 
            tag = 'div'
        return tag, rest

    def _getAtts(self, specStr):
        "gets all attr spec strings"
        if not specStr:
            return MultiDict()
        else:
            md = MultiDict()
            attKey, attVal, restStr = self._tagAttrSplit(specStr)
            md.insert(attKey, attVal)
        if restStr:
            md.update(self._getAtts(restStr))
        return md
    
    def _tagParse(self, tagStr):
        "parse a single tag"
        tagName, restStr = self._tagRestSplit(tagStr)
        attMultiDict = self._getAtts(restStr)
        return self.nodeFactory.mkElementNode(tagName, attMultiDict)

    def _decendingTagsParse(self, sameLevelStr):
        "parses a gpotentially multiple decending tags"
        tagStrL = self._tagDecendingSplit(sameLevelStr)
        tagEls = [self._tagParse(s) for s in tagStrL]
        linkNodes(tagEls)
        return tagEls[0]
    
    def _tagsParse(self, tagsStr):
        "parses a str containing potentially multiple tags"
        sameLevelStrL = self._tagSameLevelSplit(tagsStr)
        sameLevelEls = [self._decendingTagsParse(s) for s in sameLevelStrL]
        return sameLevelEls


    def _procElementTypeNodeRange(self, parentNode, lineRange):
        "processes an Element-type node"
        tagSpec, txt, isMultiLineTxt, txtStartPos = self._tagTextSplit(lineRange[0])
        sameLevelNodes = [] #initialize to avoid bomb in test below
        if isMultiLineTxt:  #mixed content is not supported
            txt = parseMultiLineTxt(txt, txtStartPos, lineRange[1:])
            lineRange = [] #it is consumed
        if tagSpec:
            sameLevelNodes = self._tagsParse(tagSpec)
            for n in sameLevelNodes:
                parentNode.addChild(n)
            txtBearingNode = sameLevelNodes[-1].lastNonTextNode()
        else:
            txtBearingNode = parentNode
        if txt:
            txtBearingNode.addChild(self.nodeFactory.mkTextNode(txt))
        if sameLevelNodes and len(lineRange) > 1:  #there's something to process
            self._procLines(sameLevelNodes[0], lineRange[1:])
    
    def _procSingleLineNodeRange(self, parentNode, lineRange):
        "processes nodes restricted to a single line"
        nodeSpec = lineRange[0].strip()
        node = self.nodeFactory.mkSingleLineNode(nodeSpec)
        parentNode.addChild(node)
        self._procLines(node, lineRange[1:])
 
    def _procSubRange(self, parentNode, lineRange):
        """updates parentNode as side effect
           this typically adds sub nodes
           processes a single sub lineRange
    
           first line defines what the content is
        """
        if not lineRange[0].strip(): #an empty line:
            self._procLines(parentNode, lineRange[1:])
        else:
            lineTypeChar = lineRange[0].lstrip()[0]
            if self.nodeFactory.isSingleLineNode(lineTypeChar):
                self._procSingleLineNodeRange(parentNode, lineRange)
            else:
                self._procElementTypeNodeRange(parentNode, lineRange)


    def _procLines(self, parentNode, lines):
        "processes lines under a given node"
        if lines:
            if not allLinesEmpty(lines):
                for lineRange in subRanges(lines):
                    self._procSubRange(parentNode, lineRange)
    
    
    def parseBuml(self, bumlStr):
        "parses a buml string and returns the root Node of a tree"
        lines = bumlStr.split('\n')
        #leave empty lines since valid in text
        #lines = [l for l in lines if l.strip()]
        lines = eliminateTabs(lines)
        #preprocess line continuations here
        rootNode = self.nodeFactory.mkRootNode()
        self._procLines(rootNode, lines)
        return rootNode

